import random
import math
import torch

class Dataloader():
    """
    Dataloader class utilized to create iterable lists of labeled tensors
    Dataloader works with pytorch tensors
    
    Each iteration on the dataloader object returns a tensor containing 
    features and labeled outputs for those features

    Warning: Make sure you input both X and Y as column vectors
    
    Usage: 
    train_data = Dataloader(X_train, Y_train, batch_size=64)
    
    for x, y in train_data:
        print("X: ", X, "Y: ", Y)
    """
    def __init__(self, X: torch.Tensor, Y: torch.Tensor, batch_size: int = 1):
        """
        inputs
        ---
        X: Tensor of features as column vectors
        Y: Tensor of Output labels as column vectors
        batch_size: number of elements per batch. Defaults to 1
        """
        self.X = X
        self.Y = Y
        self.batch_size = batch_size
        self.validate_data()
        self.n_batches = math.ceil(self.X.shape[1] / self.batch_size)
        self._iter_index = 0

    def __iter__(self):
        return self
    
    def __len__(self):
        return self.n_batches

    def __next__(self):
        position = self._iter_index * self.batch_size
        if self._iter_index >= len(self):
            # reset iter and stop current iteration
            self._iter_index = 0
            raise StopIteration
        x = self.X[:, position:position+self.batch_size]
        y = self.Y[:, position:position+self.batch_size]
        self._iter_index += 1
        return x, y

    def __repr__(self):
        return f"Dataloader(X: {self.X.shape}, Y: {self.Y.shape}, n_batches: {self.n_batches}, batch_size: {self.batch_size})"

    def validate_data(self):
        """
        Runs input data validations
        """
        assert self.X.shape[1] == self.Y.shape[1], "X and Y do not have the same number of columns/items"
        assert type(self.batch_size) == int, "batch_size should be integer value"
        assert math.floor(self.X.shape[1] / self.batch_size) >= 1, f"X tensor is not divisible by {self.batch_size}. Enter a valid batch size"
    
    def get_random_element(self):
        """
        Returns a single random element with its features and label
        """
        elem = random.randint(0, len(self)-1)
        x = self.X[:, elem].reshape(self.X.shape[0], 1)
        y = self.Y[:, elem].reshape(self.Y.shape[0], 1)
        return x, y

    def shuffle(self) -> None:
        """
        Shuffles X and Y, with features and labels pairs maintained
        """
        x = self.X.T
        y = self.Y.T
        p = np.random.permutation(len(x))
        self.X = x[p].T
        self.Y = y[p].T


import numpy as np
from abc import ABC, abstractmethod
from dataclasses import dataclass
import torch

@dataclass
class Layer(ABC):
    """
    Base layer class.

    All layers must inherit from this class
    because the NN works with Layer objects
    """
    parameters: torch.Tensor
    inputs: torch.Tensor
    outputs: torch.Tensor

    def __init__(self, parameters: dict[str, torch.Tensor] = {}) -> None:
        self.inputs = None
        self.outputs = None
        self.parameters = parameters
        self._freeze = False
        self._train = False
        
    @abstractmethod
    def forward(self, x: torch.Tensor) -> torch.Tensor:
        pass

    def get_params(self) -> dict[str, torch.Tensor]:
        """Returns a dictionary of all trainable parameters by the layer"""
        return self.parameters

    def get_train(self):
        return self._train

    def set_train(self, state):
        assert type(state) == bool
        self._train = state
    
    def set_freeze(self, state):
        """
        Freeze layer. Useful when doing transfer learning (pending implementation)
        """
        assert type(state) == bool
        self._freeze = state

class Dense(Layer):
    def __init__(self, in_size: int, out_size: int, weight_init_scale = 0.1, seed: int = 0) -> None:
        if seed:
            torch.random.manual_seed(seed)
        self.weights = torch.rand(in_size, out_size, dtype=torch.float32, requires_grad=True)
        self.bias = torch.zeros((in_size, 1), dtype=torch.float32, requires_grad=True)
        super().__init__({'weights': self.weights, 'bias': self.bias})

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.inputs = x
        self.outputs = torch.mm(self.weights, self.inputs) + self.bias
        return self.outputs

class Dropout(Layer):
    def __init__(self, drop_prob: float = 0) -> None:
        assert 0 <= drop_prob < 1, "Drop probability rate should be between 0 and 0.9"
        self.drop_prob = drop_prob
        self.keep_prob = 1 - self.drop_prob
        super().__init__()

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        self.inputs = x
        matrix_probs = [self.drop_prob, 1 - self.drop_prob]
        if self._train:
            matrix_probs = [self.drop_prob, 1 - self.drop_prob]
            self.drop_matrix = torch.from_numpy(np.random.choice([0, 1], size=x.shape, p=matrix_probs))
            self.outputs = (self.drop_matrix * self.inputs) / self.keep_prob
        else:
            self.outputs = self.inputs
        return self.outputs

    
# class BatchNormalization(Layer):
#     def __init__(self, n_rows: int = None, n_columns: int = None, weight_init_scale = 0.1, seed: int = None) -> None:
#         if seed:
#             np.random.seed(seed)
#         self.gamma = np.random.randn(n_rows, n_columns) * weight_init_scale
#         self.beta = np.random.randn(n_rows, 1)
#         self.gamma_prime = np.zeros(self.bias.shape)
#         self.beta_prime = np.zeros(self.weights.shape)
#         super().__init__()

####################
# Activation Layer #
####################

class ActivationLayer(Layer):
    def __init__(self, activation) -> None:
        self.activation = activation
        super().__init__()

    def forward(self, x) -> torch.Tensor:
        self.inputs = x
        self.outputs = self.activation(self.inputs)
        return self.outputs

class Sigmoid(ActivationLayer):
    def __init__(self) -> None:
        sigmoid = lambda x: 1 / (1 + torch.exp(-x))
        super().__init__(sigmoid)

class ReLU(ActivationLayer):
    def __init__(self) -> None:
        relu = lambda x: torch.maximum(x, torch.zeros(x.shape))
        super().__init__(relu)


class Softmax(Layer):
    def forward(self, x) -> torch.Tensor:
        self.inputs = x
        input_stable = torch.exp(x - x.max(axis=0).values)
        self.outputs = input_stable / input_stable.sum(axis=0)
        return self.outputs
        


import numpy as np
import pandas as pd
import dill
from redmind.network import NeuralNetwork

def one_hot_encode(position: int, num_classes: int):
    """
    One hot encodes a given class into a (1, num_classes) row vector
    """
    zero_vector = np.zeros((1, num_classes), dtype=int) 
    zero_vector[0][position] = 1
    # return 1d row vector
    return zero_vector.reshape(-1)

def save_model(nn: NeuralNetwork, filename: str = 'nn.dill') -> None:
    if isinstance(nn, NeuralNetwork):
        print(f"Saving Neural Network into {filename}")
        with open(filename, 'wb') as f:
            dill.dump(nn, f)
    else:
        print("Please provide a valid Neural Network")
    return None

def load_model(filename: str = 'nn.dill') -> NeuralNetwork:
    print(f"Loading NN {filename}")
    with open(filename, 'rb') as f:
        nn = dill.load(f)
    return nn

def split_dataframe(dataframe: pd.core.frame.DataFrame, y_col_idx = -1, train_percent=80, y_one_hot_encode = False, num_classes = 0, shuffle=False) -> np.ndarray:
    """
    Splits pandas dataframe into 3 groups train, dev and test sets and returns datasets as numpy arrays
    It returns the following
    X_train, Y_train / X_dev, Y_dev / X_test, Y_test
    Usage:  X_train, Y_train, X_dev, Y_dev, X_test, Y_test = split_dataframe(dataframe, y_col_idx = 5, train_percent=80)

    Arguments
    dataframe: Dataframe containing X and Y labels. dataframe is expected as row vectors
    y_col_idx: Y label column index. If no argument is passed then it will asume y label is the last column (-1)
    train_percent: percentage of data used for training set. dev and test sets will get the remaining split by 2.
                   for example if train_percent=60, then dev and test will get 20 and 20 percent respectively
    shuffle: Wether or not to shuffle the data randomly
    y_one_hot_encode: If True all Y outputs will be one hot encoded
    num_classes: If y_one_hot_encode is enabled you will need to set this to the number of classes
    """
    assert type(dataframe) == pd.core.frame.DataFrame, "Data should be entered as pandas Dataframe"
    assert train_percent > 50 and train_percent < 100, "train_percent should be between 50 and 100"
    print(f"Entered {dataframe.shape[0]} rows")
    total_samples = dataframe.shape[0]
    train_samples = int(total_samples * train_percent / 100)
    dev_samples = int((total_samples - train_samples) / 2)
    test_samples = int(total_samples - train_samples - dev_samples)
    if shuffle:
        dataframe = dataframe.sample(frac=1)
    print(f"Data split: train: {train_samples}, dev: {dev_samples}, test: {test_samples}")
    # remove Y label column first
    dataset_y = dataframe.iloc[:, y_col_idx]
    dataset_x = dataframe.drop(dataframe.columns[y_col_idx], axis=1)

    X_train = dataset_x.iloc[:train_samples].to_numpy()
    Y_train = dataset_y.iloc[:train_samples].to_numpy()

    X_dev = dataset_x.iloc[train_samples:train_samples + dev_samples].to_numpy()
    Y_dev = dataset_y.iloc[train_samples:train_samples + dev_samples].to_numpy()
    
    X_test = dataset_x.iloc[total_samples - test_samples:].to_numpy()
    Y_test = dataset_y.iloc[total_samples - test_samples:].to_numpy()
    # One hot encode Y's
    if y_one_hot_encode:
        assert num_classes > 0, "num_classes should be above 0"
        Y_train = np.array([one_hot_encode(i, num_classes) for i in Y_train])
        Y_dev = np.array([one_hot_encode(i, num_classes) for i in Y_dev])
        Y_test = np.array([one_hot_encode(i, num_classes) for i in Y_test])
    # else return Y's as column vectors
    else:
        Y_train = Y_train.reshape(Y_train.shape[0], 1)
        Y_dev = Y_dev.reshape(Y_dev.shape[0], 1)
        Y_test = Y_test.reshape(Y_test.shape[0], 1)

    return X_train, Y_train, X_dev, Y_dev, X_test, Y_test


from redmind.layers import Layer
from dataclasses import dataclass
import torch

@dataclass
class NeuralNetwork:
    layers: list[Layer]

    def __init__(self, layers: list[Layer] = [], verbose=False) -> None:
        self.layers = layers
        self._verbose = verbose
        if self._verbose:
            print(f"Neural Network initialized with {len(self.layers)} layers")

    def set_train(self, state=False):
        """
        Set network train state
        This is useful if you are using dropout layers. If so make sure properly
        changing the training states for better accuracy
        """
        if self._verbose:
            print(f"updating NN layers training to: {state}")
        for layer in self.layers:
            layer.set_train(state=state)
    
    def layers_parameters(self):
        return [p.get_params() for p in self.layers]

    def set_verbose(self, state):
        assert type(state) == bool
        self._verbose = state

    def forward(self, x: torch.Tensor = None) -> torch.Tensor:
        out = x
        for layer in self.layers:
            out = layer.forward(out)
        return out

    def predict(self, x: torch.Tensor = None) -> torch.Tensor:
        """
        Same as forward, its just used for convention
        """
        return self.forward(x)





#################################
# Learning rate decay functions #
#################################
def lr_decay(learning_rate, epoch, decay_rate):
    """Standard learning rate decay algorithm"""
    return (1 / (1 + decay_rate * epoch)) * learning_rate


from dataclasses import dataclass
from abc import ABC, abstractmethod
import torch

@dataclass
class Loss(ABC):
    value: torch.Tensor = None

    @abstractmethod
    def __call__(self, x):
        pass

class MSELoss(Loss):
    def __call__(self, y, y_pred):
        self.value =  (y_pred - y).pow(2).mean()
        return self.value
    
class CrossEntropyLoss(Loss):
    def __call__(self, y, y_pred):
        self.value =  -torch.mean(y * torch.log(y_pred))
        return self.value

class BinaryCrossEntropyLoss(Loss):
    def __call__(self, y, y_pred):
        self.value =  -torch.mean((y * torch.log(y_pred) + (1 - y) * torch.log(1 - y_pred)))
        return self.value


import torch

class Normalizer():
    """
    Class to normalize input features
    """
    def __init__(self):
        self.mean = 0
        self.variance = 0

    def fit(self, X: torch.Tensor, axis=0) -> None:
        """
        Set normalizer mean and variance to fit the data.
        If data is entered as column vector use axis 1
        IF data is entered as row vector use axis 0
        """
        self.mean = torch.mean(X, axis=axis)
        self.mean = self.mean.reshape(self.mean.shape[0], 1)
        self.std = torch.std(X - self.mean, axis=axis)
        self.std = self.std.reshape(self.std.shape[0], 1)

    def scale(self, X: torch.Tensor):
        """
        Returns normalized input
        """
        return (X - self.mean) / self.std




"""
Neural network optimizers
"""
from abc import ABC, abstractmethod
import torch

def init_velocity_vector(layers: list[dict[str, torch.Tensor]]) -> list[dict[str, torch.Tensor]]:
    velocity_layers_list = []
    # build layers velocity dict with np zeros array for each trainable param
    for layer in layers:
        velocity_map = {}
        for param_name, param_value in layer.items():
            velocity_map[param_name] = torch.zeros(param_value.shape)
        velocity_layers_list.append(velocity_map)
    return velocity_layers_list

class Optimizer(ABC):
    """
    Only one optimizer can be assigned to the entire NN

    The optimizer is in charge of keeping track of optimization
    variable states for all layers. Check Adam optimizer for reference.

    Optimizer workflow
    1. Loop through self.params (coming from Neural Network layers)
    3. Optimize those parameters according to strategy
    4. (optional) track optimization
    5. Update layer parameters on learning rate scale
    """
    def __init__(self, layers_params: list[dict[str, torch.Tensor]], learning_rate: float = 1e-2):
        #assert isinstance(network, NeuralNetwork), "network should be a NeuralNetwork object"
        self.params = layers_params
        self.learning_rate = learning_rate
    
    def zero_grad(self) -> None:
        """Sets all parameters gradients to zero"""
        for layer in self.params:
            for _, param_value in layer.items():
                param_value.grad = torch.zeros(param_value.shape)
        
    @abstractmethod
    def step(self) -> None:
        """Runs optimizer for all trainable parameters returned by the layer"""
        
class GradientDescent(Optimizer):
    def step(self) -> None:
        for layer in self.params:
            for param_name, param_value in layer.items():
                direction = param_value.grad * self.learning_rate
                # Running in place operation with context manager
                # because we dont need to track this operation
                with torch.no_grad():
                    layer[param_name] -= direction

class Momentum(Optimizer):
    beta: float

    def __init__(self, layers_params: list[dict[str, torch.Tensor]], learning_rate: float = 1e-2, beta = 0.9):
        super().__init__(layers_params, learning_rate)
        self.beta = beta
        self.gradients_velocity = init_velocity_vector(self.params)


    def step(self) -> None:
        for idx, layer in enumerate(self.params):
            for param_name, param_value in layer.items():
                self.gradients_velocity[idx][param_name] = self.beta * self.gradients_velocity[idx][param_name] + (1 - self.beta) * param_value.grad
                direction = self.gradients_velocity[idx][param_name] * self.learning_rate
                # Running in place operation with context manager
                # because we dont need to track this operation
                with torch.no_grad():
                    layer[param_name] -= direction

class RMSprop(Optimizer):
    beta: float
    epsilon: float = 1e-7

    def __init__(self, layers_params: list[dict[str, torch.Tensor]], learning_rate: float = 1e-2, beta = 0.9):
        super().__init__(layers_params, learning_rate)
        self.beta = beta
        self.gradients_velocity = init_velocity_vector(self.params)

    
    def step(self) -> None:
        for idx, layer in enumerate(self.params):
            for param_name, param_value in layer.items():
                self.gradients_velocity[idx][param_name] = self.beta * self.gradients_velocity[idx][param_name] + (1 - self.beta) * torch.pow(param_value.grad, 2)
                direction = param_value.grad / torch.sqrt(self.gradients_velocity[idx][param_name] + self.epsilon) * self.learning_rate
                # Running in place operation with context manager
                # because we dont need to track this operation
                with torch.no_grad():
                    layer[param_name] -= direction


### Pending to fix
class Adam(Optimizer):
    """Adam is a combination of momentum and RMSprop, thats why the velocity names"""
    beta1 = 0.9
    beta2 = 0.999
    epsilon = 1e-7

    def __init__(self, layers_params: list[dict[str, torch.Tensor]], learning_rate = 1e-2):
        super().__init__(layers_params, learning_rate)
        self.momentum_velocity = init_velocity_vector(self.params)
        self.rmsprop_velocity = init_velocity_vector(self.params)

    def step(self) -> None:
        for idx, layer in enumerate(self.params):
            for param_name, param_value in layer.items():
                self.momentum_velocity[idx][param_name] = (self.beta1 * self.momentum_velocity[idx][param_name]) + ((1 - self.beta1) * param_value.grad)
                self.rmsprop_velocity[idx][param_name] = (self.beta2 * self.rmsprop_velocity[idx][param_name]) + ((1 - self.beta2) * torch.pow(param_value.grad, 2))
                direction = (self.momentum_velocity[idx][param_name] / torch.sqrt(self.rmsprop_velocity[idx][param_name] + self.epsilon)) * self.learning_rate
                # Running in place operation with context manager
                # because we dont need to track this operation
                with torch.no_grad():
                    layer[param_name] -= direction




from setuptools import setup, find_packages
import codecs
import os

here = os.path.abspath(os.path.dirname(__file__))

with codecs.open(os.path.join(here, "README.md"), encoding="utf-8") as fh:
    long_description = "\n" + fh.read()

# This gets deployed when a new release is made by github actions
VERSION = '0.0.9'
DESCRIPTION = 'Package for building Machine learning models'
LONG_DESCRIPTION = 'A package that allows you to build complex machine learning models in a simple fashion'

setup(
    name="redmind",
    version=VERSION,
    author="Diego Velez",
    author_email="diegovelezs94@gmail.com",
    description=DESCRIPTION,
    long_description_content_type="text/markdown",
    long_description=long_description,
    packages=find_packages(),
    install_requires=['numpy', 'matplotlib', 'dill', 'pandas', 'torch'],
    keywords=['python', 'ai', 'machine learning', 'neural network'],
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ]
)


import numpy as np
import matplotlib.pyplot as plt
import redmind.functions as fn
from redmind.network import NeuralNetwork
from redmind.layers import Layer
from redmind.dataloader import Dataloader
from redmind.optimizers import Optimizer
from redmind.loss import Loss
import torch


class Trainer():
    """
    Trainer class simply helps by automating common training process. 
    It has predefined trainig functions so you only need to send the arguments. 
    
    Using this class is not required, however its very convenient.
    """
    def __init__(self, network: NeuralNetwork, loss_function: Loss, optimizer: Optimizer):
        assert isinstance(network, NeuralNetwork), "network should be a NeuralNetwork object"
        assert isinstance(optimizer, Optimizer), "optimizer should be an Optimizer object"
        assert isinstance(loss_function, Loss), "loss_function should be a Loss object"
        self.network = network
        self.optimizer = optimizer
        self.costs = {}
        self.loss_function = loss_function
    
    def train(self, X: torch.Tensor, Y: torch.Tensor, epochs: int = 100, batch_size: int = 1, early_stoping: float = 0.0):
        assert type(X) == torch.Tensor, "expecting X to be a torch tensor"
        assert type(Y) == torch.Tensor, "expecting Y to be a torch tensor"

        data = Dataloader(X=X, Y=Y, batch_size=batch_size)
        self.network.set_train(state=True)

        for epoch in range(epochs):
            epoch_losses = []
            for x, y in data:
                # forward
                y_pred = self.network.forward(x)
                # clear gradients
                self.optimizer.zero_grad()

                # calculate error and cost
                loss = self.loss_function(y, y_pred)
                epoch_losses.append(loss.detach())
                loss.backward()
                
                # Gradient descent step
                self.optimizer.step()

            # print cost to console
            self.costs[epoch] = torch.stack(epoch_losses).mean().item()
            accuracy = round(100 - (self.costs[epoch] * 100), 3)
            print(f"epoch: {epoch + 1}/{epochs}, cost: {round(self.costs[epoch], 4)}, accuracy: {accuracy}%")

            # Early stoping
            if early_stoping > 0.0 and accuracy >= early_stoping:
                print(f"Early stoping threshold reached (over {early_stoping}%)")
                break

        self.network.set_train(state=False)

    def graph_costs(self) -> None:
        plt.plot(list(self.costs.keys()), list(self.costs.values()))
        plt.xlabel("Epoch")
        plt.ylabel("Cost")
        plt.show()